
KM_NULL_STATE	= 0;
KM_PREQUERY		= 1;
KM_INQUERY		= 2;
KM_POSTQUERY	= 3;
KM_ANALYZING	= 4;

local AUCTION_CLASS_WEAPON = 1;
local AUCTION_CLASS_ARMOR  = 2;

local gAllScans = {};

local BIGNUM = 999999999999;

-----------------------------------------

AtrScan = {};
AtrScan.__index = AtrScan;


-----------------------------------------

local function ItemType2AuctionClass(itemType)
	local itemClasses = { GetAuctionItemClasses() };
	if #itemClasses > 0 then
	local itemClass;
		for x, itemClass in pairs(itemClasses) do
			if (itemClass == itemType) then
				return x;
			end
		end
	end
end


-----------------------------------------

local function SubType2AuctionSubclass(auctionClass, itemSubtype)
	local itemClasses = { GetAuctionItemSubClasses(auctionClass) };
	if #itemClasses > 0 then
	local itemClass;
		for x, itemClass in pairs(itemClasses) do
			if (itemClass == itemSubtype) then
				return x;
			end
		end
	end
end

-----------------------------------------

function Atr_FindScan (itemName, itemLink)

	if (itemName == nil or itemName == "") then
		itemName = "nil";
	end

	local itemNameLC = string.lower (itemName);

	if (gAllScans[itemNameLC] == nil) then

		local scn = {};
		setmetatable (scn, AtrScan);

		scn:Init ();

		scn.itemName			= itemName;
		scn.itemLink			= itemLink;
		scn.itemSoftName		= nil;
		scn.itemClass			= 0;
		scn.itemSubclass		= 0;


		-- get the class and subclass for faster AH queries and see if this is an item in a "family"

		if (itemLink) then
	--		local printable = gsub(gCurrentItemLink, "\124", "\124\124");
			local a, b, c, d, e, sType, sSubType = GetItemInfo(itemLink);

			scn.itemClass		= ItemType2AuctionClass (sType);
			scn.itemSubclass	= SubType2AuctionSubclass (scn.itemClass, sSubType);

			if (scn.itemClass == AUCTION_CLASS_WEAPON or scn.itemClass == AUCTION_CLASS_ARMOR) then
				local z = string.find (string.lower(scn.itemName), " of ", 1, true);
				if (z ~= nil) then
					scn.itemSoftName = string.sub (scn.itemName, 1, z-1);
					if (not ZC_StringContains (scn.itemSoftName, " ")) then
						scn.itemSoftName = nil;
					end
				end
			end
		end

--		ZC_chatmsg ("Adding "..itemNameLC.." to cache", .4, .4, 1);
		
		gAllScans[itemNameLC] = scn;
	end
	
	return gAllScans[itemNameLC];
end

-----------------------------------------

function Atr_ClearScanCache ()

--	ZC_redmsg ("Clearing Scan Cache");

	for a,v in pairs (gAllScans) do
		if (a ~= "nil") then
			gAllScans[a] = nil;
		end
	end

end

-----------------------------------------

function AtrScan:Init ()
	self.processing_state	= KM_NULL_STATE
	self.current_page		= -1
	self.scanData			= {};
	self.sortedData			= {};
	self.whenScanned		= 0;
	self.whenSorted			= 0;
	self.lowprices			= {BIGNUM, BIGNUM, BIGNUM};
end

-----------------------------------------

function AtrScan:Start ()

	if (self.itemName == "nil") then
		return;
	end

	BrowseName:SetText (self.itemName);		-- not necessary but nice when user switches to Browse tab
	
	SortAuctionItems("list", "buyout");

	if (IsAuctionSortReversed("list", "buyout")) then
		SortAuctionItems("list", "buyout");
	end

	self:Init ();

	self.current_page		= 0;
	self.processing_state	= KM_PREQUERY;

	self:Continue();
end

-----------------------------------------

function AtrScan:AnalyzeResultsPage()

	self.processing_state = KM_ANALYZING;

	local numBatchAuctions, totalAuctions = GetNumAuctionItems("list");

--	ZC_chatmsg("auctions:"..numBatchAuctions.." out of  "..totalAuctions)

	if (totalAuctions >= 50) then
		Atr_SetMessage ("Scanning auctions: page "..self.current_page);
	end

	if (numBatchAuctions > 0) then

		local x;

		for x = 1, numBatchAuctions do

			local name, texture, count, quality, canUse, level, minBid, minIncrement, buyoutPrice, bidAmount, highBidder, owner = GetAuctionItemInfo("list", x);

			local exactMatch = ZC_StringSame (name, self.itemName);

			if (exactMatch or ZC_StringStartsWith (name, self.itemSoftName)) then
				self:AddScanItem (name, count, buyoutPrice, owner);
			
				if (self.itemLink == nil) then
					self.itemLink = GetAuctionItemLink("list", x);
				end
			end
			
			if (exactMatch) then
				self.itemName = name;		-- just cleans up any case differences
			end
			
		end
	end

	local done = (numBatchAuctions < 50);

	if (not done) then
		self.processing_state = KM_PREQUERY;
	end;

	return done;
end

-----------------------------------------

function AtrScan:Get_Suffix (name)

	if (self.itemSoftName) then
		return string.sub (name, strlen (self.itemSoftName) + 2);
	end

	return "";
end

-----------------------------------------

function AtrScan:AddScanItem (name, stackSize, buyoutPrice, owner, numAuctions)

	local sd = {};
	local i;

	if (numAuctions == nil) then
		numAuctions = 1;
	end

	for i = 1, numAuctions do
		sd["stackSize"]		= stackSize;
		sd["buyoutPrice"]	= buyoutPrice;
		sd["owner"]			= owner;
		sd["pagenum"]		= self.current_page - 1;
		sd["suffix"]		= self:Get_Suffix(name);

		tinsert (self.scanData, sd);
		
		local itemPrice = math.floor (buyoutPrice / stackSize);

		Atr_AddToLowPrices (self.lowprices, itemPrice);
	end

end


-----------------------------------------

function AtrScan:AddSDXToScan (price, owner, volume)	-- helper function for AddExternalDataToScan

	local sd = {};

	if (price and price > 0) then
		sd["stackSize"]		= 1;
		sd["buyoutPrice"]	= price;
		sd["owner"]			= owner;
		sd["pagenum"]		= -1;
		sd["suffix"]		= "";

		if (volume) then
			sd["volume"] = volume;
		end

		tinsert (self.scanData, sd);
	end
	
end

-----------------------------------------

function AtrScan:AddExternalDataToScan ()

	if (self.itemLink == nil) then
		return;
	end

	-- Wowecon

	if (Wowecon and Wowecon.API) then
	
		local priceG, volG = Wowecon.API.GetAuctionPrice_ByLink (self.itemLink, Wowecon.API.GLOBAL_PRICE)
		local priceS, volS = Wowecon.API.GetAuctionPrice_ByLink (self.itemLink, Wowecon.API.SERVER_PRICE)

		self:AddSDXToScan (priceG, "__wowEconG", volG);
		self:AddSDXToScan (priceS, "__wowEconS", volS);
		
	end
	
	-- GoingPrice Wowhead
	
	local id = ZC_ItemIDfromLink (self.itemLink);
	
	id = tonumber(id);

	if (GoingPrice_Wowhead_Data and GoingPrice_Wowhead_Data[id] and GoingPrice_Wowhead_SV._index) then
		local index = GoingPrice_Wowhead_SV._index["Buyout price"];

		if (index ~= nil) then
			local price = GoingPrice_Wowhead_Data[id][index];
		
			self:AddSDXToScan (price, "__wowHead");
		end
	end

	-- GoingPrice Allakhazam
	
	if (GoingPrice_Allakhazam_Data and GoingPrice_Allakhazam_Data[id] and GoingPrice_Allakhazam_SV._index) then
		local index = GoingPrice_Allakhazam_SV._index["Median"];

		if (index ~= nil) then
			local price = GoingPrice_Allakhazam_Data[id][index];
		
			self:AddSDXToScan (price, "__allakhazam");
		end
	end

	-- most recent historical price
	
	local price = Atr_Process_Historydata();
	if (price ~= nil) then
		self:AddSDXToScan (price, "__atrLast");
	end

end

-----------------------------------------

function AtrScan:SubtractScanItem (name, stackSize, buyoutPrice)

	local sd;
	local i;

	for i,sd in ipairs (self.scanData) do
		
		if (sd.stackSize == stackSize and sd.buyoutPrice == buyoutPrice and ZC_StringSame (sd.suffix, self:Get_Suffix(name))) then
			
			tremove (self.scanData, i);
			return;
		end
	end

end

-----------------------------------------

function AtrScan:Continue()

	if (CanSendAuctionQuery()) then

		self.processing_state = KM_IN_QUERY;

		local queryString = self.itemName;

		if (self.itemSoftName) then
			queryString = self.itemSoftName;
		end

		queryString = string.sub (queryString,0,31);	-- attempting to reduce number of disconnects

--		ZC_chatmsg ("AtrScan:Continue   "..queryString.."  page:"..self.current_page);
		
		QueryAuctionItems (queryString, "", "", nil, self.itemClass, self.itemSubclass, self.current_page, nil, nil);

		self.processing_state = KM_POSTQUERY;
		self.current_page = self.current_page + 1;
	end

end

-----------------------------------------

function AtrScan:Finish()

	self.whenScanned		= time();
	self.processing_state	= KM_NULL_STATE;
	self.current_page		= -1;
	
	self:CondenseAndSort ();

	-- update the scan DB
	
	local newprice = Atr_CalcNewDBprice (self.itemName, self.lowprices);
	
	if (newprice > 0) then
		AUCTIONATOR_PRICE_DATABASE[self.itemName] = newprice;
	end

end

-----------------------------------------

function Atr_SortAuctionData (x, y)

	return x.itemPrice < y.itemPrice;

end

-----------------------------------------

function AtrScan:CondenseAndSort ()

	----- Condense the scan data into a table that has only a single entry per stacksize/price combo

	self.sortedData	= {};
	self.whenSorted	= time();

	local i,sd;
	local conddata = {};

	for i,sd in ipairs (self.scanData) do

		local ownerCode = "x";
		local dataType  = "n";		-- normal
		
		if (sd.owner == UnitName("player")) then
			ownerCode = "y";
		elseif (sd.owner == "__wowEconG") then
			dataType = "eg";
		elseif (sd.owner == "__wowEconS") then
			dataType = "es";
		elseif (sd.owner == "__wowHead") then
			dataType = "h";
		elseif (sd.owner == "__allakhazam") then
			dataType = "k";
		elseif (sd.owner == "__atrLast") then
			dataType = "a";
		end

		local key = "_"..sd.stackSize.."_"..sd.buyoutPrice.."_"..ownerCode..dataType.."_"..sd.suffix;

		if (conddata[key] and gSoftMatchItemName == nil) then
			conddata[key].count = conddata[key].count + 1;
		else
			local data = {};

			data.stackSize 		= sd.stackSize;
			data.buyoutPrice	= sd.buyoutPrice;
			data.itemPrice		= sd.buyoutPrice / sd.stackSize;
			data.suffix			= sd.suffix;
			data.pagenum		= sd.pagenum;
			data.count			= 1;
			data.type			= dataType;
			data.yours			= (ownerCode == "y");
			
			if (sd.volume) then
				data.volume = sd.volume;
			end
			
			conddata[key] = data;
		end

	end

	----- create a table of these entries

	local n = 1;

	local i, v;

	for i,v in pairs (conddata) do
		self.sortedData[n] = v;
		n = n + 1;
	end

	-- sort the table by itemPrice

	table.sort (self.sortedData, Atr_SortAuctionData);

	-- analyze and store some info about the data

	self:AnalyzeSortData ();


end

-----------------------------------------

function AtrScan:AnalyzeSortData ()

	self.absoluteBest			= nil;
	self.bestPrices				= {};		-- a table with one entry per stacksize that is the cheapest auction for that particular stacksize
	self.numMatches				= 0;
	self.numMatchesWithBuyout	= 0;
	self.hasStack				= false;

	local j, sd;

	----- find the best price per stacksize and overall -----

	for j,sd in ipairs(self.sortedData) do

		if (sd.type == "n") then

			if (self.itemSoftName == nil or self.itemName == self.itemSoftName.." "..sd.suffix) then

				self.numMatches = self.numMatches + 1;

				if (sd.itemPrice > 0) then

					self.numMatchesWithBuyout = self.numMatchesWithBuyout + 1;

					if (self.bestPrices[sd.stackSize] == nil or self.bestPrices[sd.stackSize].itemPrice >= sd.itemPrice) then
						self.bestPrices[sd.stackSize] = sd;
					end

					if (self.absoluteBest == nil or self.absoluteBest.itemPrice > sd.itemPrice) then
						self.absoluteBest = sd;
					end
				end

				if (sd.stackSize > 1) then
					self.hasStack = true;
				end
			end
		end
	end
end

-----------------------------------------

function AtrScan:FindInSortedData (stackSize, buyoutPrice)
	local j = 1;
	for j = 1,#self.sortedData do
		sd = self.sortedData[j];
		if (sd.stackSize == stackSize and sd.buyoutPrice == buyoutPrice and sd.yours) then
			return j;
		end
	end
	
	return 0;
end


-----------------------------------------

function AtrScan:FindMatchByStackSize (stackSize)

	local index = nil;

	local basedata = self.absoluteBest;

	if (self.bestPrices[stackSize]) then
		basedata = self.bestPrices[stackSize];
	end

	local numrows = #self.sortedData;

	local n;

	for n = 1,numrows do

		local data = self.sortedData[n];

		if (basedata and data.itemPrice == basedata.itemPrice and data.stackSize == basedata.stackSize and data.suffix == basedata.suffix and data.yours == basedata.yours) then
			index = n;
			break;
		end
	end

	return index;
	
end

-----------------------------------------

function AtrScan:FindMatchByYours ()

	local index = nil;

	local j;
	for j = 1,#self.sortedData do
		sd = self.sortedData[j];
		if (sd.yours) then
			index = j;
			break;
		end
	end

	return index;

end

-----------------------------------------

function AtrScan:FindCheapest ()

	local index = nil;

	local j;
	for j = 1,#self.sortedData do
		sd = self.sortedData[j];
		if (sd.itemPrice > 0) then
			index = j;
			break;
		end
	end

	return index;

end



-----------------------------------------

ATR_FS_NULL			= 0;
ATR_FS_STARTED		= 1;
ATR_FS_ANALYZING	= 2;
ATR_FS_CLEANING_UP	= 3;

gAtr_FullScanState = ATR_FS_NULL;

local Atr_orig_WOWEcon_Scan_AH;

-----------------------------------------

function Atr_GetDBsize()

	local n = 0;
	local a,v;

	for a,v in pairs (AUCTIONATOR_PRICE_DATABASE) do
		n = n + 1;
	end
	
	return n;
end

-----------------------------------------

function Atr_WOWEcon_Scan_AH()

	ZC_redmsg ("Atr_WOWEcon_Scan_AH");

end

-----------------------------------------

function Atr_FullScanStart()

	local canQuery,canQueryAll = CanSendAuctionQuery();
	
	if (canQueryAll) then
	
		if (WOWEcon_Scan_AH) then
			Atr_orig_WOWEcon_Scan_AH = WOWEcon_Scan_AH;
			WOWEcon_Scan_AH = Atr_WOWEcon_Scan_AH;
		end
	
		Atr_FullScanStatus:SetText ("Scanning...");
		Atr_FullScanStartButton:Disable();
		Atr_FullScanDone:Disable();
	
		gAtr_FullScanState = ATR_FS_STARTED;
		QueryAuctionItems ("", nil, nil, 0, 0, 0, 0, 0, 0, true);
	end

end

-----------------------------------------

function Atr_CalcNewDBprice (name, prices)
		
	if (AUCTIONATOR_PRICE_DATABASE[name]) then
		prices[3] = AUCTIONATOR_PRICE_DATABASE[name];
	end
	
	table.sort (prices);

	if (prices[2] ~= BIGNUM and prices[1] < (prices[2]/2)) then
		prices[1] = BIGNUM;
	end
	
	local total = 0;
	local n = 0;
	local k;
	
	for k = 1,3 do
--		ZC_yellowmsg (prices[k]);
		if (prices[k] ~= BIGNUM) then
			total = total + prices[k];
			n = n + 1;
		end
	end

	if (n == 0) then return 0; end;
	
	return math.floor (total / n);
end

-----------------------------------------

function Atr_AddToLowPrices (lowprices, itemPrice)
	
	if (itemPrice > 0) then
		if (itemPrice < lowprices[1]) then
			lowprices[1] = itemPrice;
		elseif (itemPrice < lowprices[2]) then
			lowprices[2] = itemPrice;
		end
	end
end



-----------------------------------------

function Atr_FullScanAnalyze()

	gAtr_FullScanState = ATR_FS_ANALYZING;

	Atr_FullScanStatus:SetText ("Processing");
	
	local numBatchAuctions, totalAuctions = GetNumAuctionItems("list");

--	ZC_chatmsg("FULL SCAN:"..numBatchAuctions.." out of  "..totalAuctions)

	local lowprices = {};
	
	if (numBatchAuctions > 0) then

		local x;

		for x = 1, numBatchAuctions do

			local name, texture, count, quality, canUse, level, minBid, minIncrement, buyoutPrice = GetAuctionItemInfo("list", x);

			if (name ~= nil and buyoutPrice ~= nil) then
			
				local itemPrice = math.floor (buyoutPrice / count);
			
				if (itemPrice > 0) then
					if (not lowprices[name]) then
						lowprices[name] = {BIGNUM,BIGNUM,BIGNUM};		-- one extra for later
					end
					
					Atr_AddToLowPrices (lowprices[name], itemPrice);
				end
			end
		end
	end

	for name,prices in pairs (lowprices) do
		
		local newprice = Atr_CalcNewDBprice (name, prices);
		
		if (newprice > 0) then
			AUCTIONATOR_PRICE_DATABASE[name] = newprice;
		end
	end


	gAtr_FullScanState = ATR_FS_CLEANING_UP;

	Atr_FullScanStartButton:Enable();
	Atr_FullScanDone:Enable();
	Atr_FullScanStatus:SetText ("");
	
	AUCTIONATOR_LAST_SCAN_TIME = time();
	
	Atr_UpdateFullScanFrame ();

	QueryAuctionItems("xxzzy", 80, 80, 0, 7, 0);		-- clear out the list (seems safer)
	
	lowprices = {};
	collectgarbage ("collect");
end

-----------------------------------------

local ggg = 0;

function auctionator_AuctionFrameBrowse_Update ()

	ggg = ggg + 1;
	ZC_printstack (ggg.." ");

	return auctionator_orig_AuctionFrameBrowse_Update ();

end

-----------------------------------------

function Atr_ShowFullScanFrame()

	Atr_FullScanFrame:Show();
	Atr_FullScanFrame:SetBackdropColor(0,0,0,100);
	
	Atr_UpdateFullScanFrame();
	Atr_FullScanStatus:SetText ("");

	local tooltipWarning = "";
	if (AUCTIONATOR_SHOW_TIPS ~= 1) then
		tooltipWarning = "<br/><br/>You currently have Auctionator configured to NOT show prices in tooltips.  You can enable this option in the Basic Options window.";
	end


	local expText = "<html><body>"
					.."<p>"
					.."Scanning is entirely optional."
					.."<br/><br/>"
					.."Scanning the auction house builds a database of prices that Auctionator uses for two purposes: to display in item tooltips "
					.."when you're away from the Auction House and to recommend a price when you're posting an item for which there are no other "
					.."auctions at that moment."
					.."<br/><br/>"
					.."Scanning usually takes about 10 seconds but may take longer if your server is very busy.  Note that Blizzard only allows "
					.."scanning once every 15 minutes."
					..tooltipWarning
					.."<br/><br/>"
					.."Finally, a warning: if you have a low-bandwidth connection, scanning may cause you to disconnect."
					.."</p>"
					.."</body></html>"
					;



	Atr_FullScanHTML:SetText (expText);
	Atr_FullScanHTML:SetSpacing (3);
end

-----------------------------------------

function Atr_UpdateFullScanFrame()

	Atr_FullScanDBsize:SetText (Atr_GetDBsize());
	
	if (AUCTIONATOR_LAST_SCAN_TIME) then
		Atr_FullScanDBwhen:SetText (date ("%A, %B %d at %I:%M %p", AUCTIONATOR_LAST_SCAN_TIME));
	else
		Atr_FullScanDBwhen:SetText ("never");
	end

	local canQuery,canQueryAll = CanSendAuctionQuery();

	if (canQueryAll) then
		Atr_FullScanStatus:SetText ("");
		Atr_FullScanStartButton:Enable();
		Atr_FullScanNext:SetText("Now");
	else	
		Atr_FullScanStartButton:Disable();

		if (AUCTIONATOR_LAST_SCAN_TIME) then
			local when = 15*60 - (time() - AUCTIONATOR_LAST_SCAN_TIME);
		
			when = math.floor (when/60);
		
			if (when == 0) then
				Atr_FullScanNext:SetText ("in less than a minute");
			elseif (when == 1) then
				Atr_FullScanNext:SetText ("in about one minute");
			else
				Atr_FullScanNext:SetText ("in about "..when.." minutes");
			end
		else
			Atr_FullScanNext:SetText ("Unknown");
		end
	end
end

-----------------------------------------

function Atr_FullScanFrameIdle()

	if (gAtr_FullScanState == ATR_FS_CLEANING_UP) then
	
		if (GetNumAuctionItems("list") < 100) then
		
			if (Atr_orig_WOWEcon_Scan_AH) then
				WOWEcon_Scan_AH = Atr_orig_WOWEcon_Scan_AH;
				Atr_orig_WOWEcon_Scan_AH = nil;
			end
		
			Atr_FullScanStatus:SetText ("Scan complete");
			PlaySound("AuctionWindowClose");
			
			gAtr_FullScanState = ATR_FS_NULL;
		end
	
	end
	
	if (gAtr_FullScanState == ATR_FS_STARTED) then

		local btext = Atr_FullScanStatus:GetText ();
		
		if (btext) then
			if (string.len (btext) > 25) then
				Atr_FullScanStatus:SetText ("Scanning.");
			else
				Atr_FullScanStatus:SetText (btext..".");
			end
		end
	end
	
end



